# -*- coding: utf-8 -*-

import jax
import time
import jax.numpy as jnp
from jax._src.api import (_check_input_dtype_jacfwd, _check_input_dtype_jacrev, _check_output_dtype_jacfwd, _check_output_dtype_jacrev, _ensure_index, _jvp,
                          _vjp, _std_basis, _jacfwd_unravel, _jacrev_unravel, lu, argnums_partial, tree_map, tree_structure, tree_transpose, partial, Callable, Sequence, Union, vmap)
# fix import location for jax 0.4.1
try:
    from jax._src.api import _check_callable
except ImportError:
    from jax._src.api_util import check_callable as _check_callable


def amax(x, return_arg=False):
    """Return the maximum absolute value.
    """
    absx = jnp.abs(x)
    if return_arg:
        arg = jnp.argmax(absx)
        return absx[arg], arg
    else:
        return absx.max()


def jvp_vmap(fun: Callable, argnums: Union[int, Sequence[int]] = 0):
    """Vectorized (forward-mode) jacobian-vector product of ``fun``. This is by large adopted from the implementation of jacfwd in jax._src.api.

    Args:
      fun: Function whose value and Jacobian is to be computed.
      argnums: Optional, integer or sequence of integers. Specifies which positional argument(s) to differentiate with respect to (default ``0``).

    Returns:
      A function with the same arguments as ``fun``, that evaluates the value and Jacobian of ``fun`` using forward-mode automatic differentiation.
    """
    _check_callable(fun)
    argnums = _ensure_index(argnums)

    def jvpfun(args, tangents, **kwargs):

        f = lu.wrap_init(fun, kwargs)
        f_partial, dyn_args = argnums_partial(
            f, argnums, args, require_static_args_hashable=False)
        pushfwd = partial(_jvp, f_partial, dyn_args)
        y, jac = vmap(pushfwd, out_axes=(None, -1), in_axes=-1)(tangents)

        return y, jac

    return jvpfun


def vjp_vmap(fun: Callable, argnums: Union[int, Sequence[int]] = 0):
    """Vectorized (reverse-mode) vector-jacobian product of ``fun``. This is by large adopted from the implementation of jacrev in jax._src.api.

    Args:
      fun: Function whose value and Jacobian are to be computed.
      argnums: Optional, integer or sequence of integers. Specifies which
        positional argument(s) to differentiate with respect to (default ``0``).

    Returns:
      A function with the same arguments as ``fun``, that evaluates the value and Jacobian of
      ``fun`` using reverse-mode automatic differentiation.
    """
    _check_callable(fun)

    def vjpfun(args, tangents, **kwargs):
        f = lu.wrap_init(fun, kwargs)
        f_partial, dyn_args = argnums_partial(f, argnums, args,
                                              require_static_args_hashable=False)
        y, pullback = _vjp(f_partial, *dyn_args)
        jac = vmap(pullback)(tangents)
        return y, jac

    return vjpfun


def val_and_jacfwd(fun: Callable, argnums: Union[int, Sequence[int]] = 0,
                   has_aux: bool = False, holomorphic: bool = False) -> Callable:
    """Value and Jacobian of ``fun`` evaluated column-by-column using forward-mode AD. Apart from returning the function value, this is one-to-one adopted from jax._src.api.

    Args:
      fun: Function whose value and Jacobian are to be computed.
      argnums: Optional, integer or sequence of integers. Specifies which
        positional argument(s) to differentiate with respect to (default ``0``).
      has_aux: Optional, bool. Indicates whether ``fun`` returns a pair where the
        first element is considered the output of the mathematical function to be
        differentiated and the second element is auxiliary data. Default False.
      holomorphic: Optional, bool. Indicates whether ``fun`` is promised to be
        holomorphic. Default False.
      allow_int: Optional, bool. Whether to allow differentiating with
        respect to integer valued inputs. The gradient of an integer input will
        have a trivial vector-space dtype (float0). Default False.

    Returns:
      A function with the same arguments as ``fun``, that evaluates the value and Jacobian of
      ``fun`` using reverse-mode automatic differentiation. If ``has_aux`` is True
      then a pair of (jacobian, auxiliary_data) is returned.
    """
    _check_callable(fun)
    argnums = _ensure_index(argnums)

    def jacfun(*args, **kwargs):
        f = lu.wrap_init(fun, kwargs)
        f_partial, dyn_args = argnums_partial(f, argnums, args,
                                              require_static_args_hashable=False)
        tree_map(partial(_check_input_dtype_jacfwd, holomorphic), dyn_args)
        if not has_aux:
            pushfwd = partial(_jvp, f_partial, dyn_args)
            y, jac = vmap(pushfwd, out_axes=(None, -1))(_std_basis(dyn_args))
        else:
            pushfwd = partial(_jvp, f_partial, dyn_args, has_aux=True)
            y, jac, aux = vmap(pushfwd, out_axes=(
                None, -1, None))(_std_basis(dyn_args))
        tree_map(partial(_check_output_dtype_jacfwd, holomorphic), y)
        example_args = dyn_args[0] if isinstance(argnums, int) else dyn_args
        jac_tree = tree_map(partial(_jacfwd_unravel, example_args), y, jac)
        if not has_aux:
            return y, jac_tree
        else:
            return y, jac_tree, aux

    return jacfun


def val_and_jacrev(fun: Callable, argnums: Union[int, Sequence[int]] = 0,
                   has_aux: bool = False, holomorphic: bool = False, allow_int: bool = False) -> Callable:
    """Value and Jacobian of ``fun`` evaluated row-by-row using reverse-mode AD. Apart from returning the function value, this is one-to-one adopted from jax._src.api.

    Args:
      fun: Function whose value and Jacobian are to be computed.
      argnums: Optional, integer or sequence of integers. Specifies which
        positional argument(s) to differentiate with respect to (default ``0``).
      has_aux: Optional, bool. Indicates whether ``fun`` returns a pair where the
        first element is considered the output of the mathematical function to be
        differentiated and the second element is auxiliary data. Default False.
      holomorphic: Optional, bool. Indicates whether ``fun`` is promised to be
        holomorphic. Default False.
      allow_int: Optional, bool. Whether to allow differentiating with
        respect to integer valued inputs. The gradient of an integer input will
        have a trivial vector-space dtype (float0). Default False.

    Returns:
      A function with the same arguments as ``fun``, that evaluates the value and Jacobian of
      ``fun`` using reverse-mode automatic differentiation. If ``has_aux`` is True
      then a pair of (jacobian, auxiliary_data) is returned.
    """
    _check_callable(fun)

    def jacfun(*args, **kwargs):
        f = lu.wrap_init(fun, kwargs)
        f_partial, dyn_args = argnums_partial(f, argnums, args,
                                              require_static_args_hashable=False)
        tree_map(partial(_check_input_dtype_jacrev,
                 holomorphic, allow_int), dyn_args)
        if not has_aux:
            y, pullback = _vjp(f_partial, *dyn_args)
        else:
            y, pullback, aux = _vjp(f_partial, *dyn_args, has_aux=True)
        tree_map(partial(_check_output_dtype_jacrev, holomorphic), y)
        jac = vmap(pullback)(_std_basis(y))
        jac = jac[0] if isinstance(argnums, int) else jac
        example_args = dyn_args[0] if isinstance(argnums, int) else dyn_args
        jac_tree = tree_map(partial(_jacrev_unravel, y), example_args, jac)
        jac_tree = tree_transpose(tree_structure(
            example_args), tree_structure(y), jac_tree)
        if not has_aux:
            return y, jac_tree
        else:
            return y, jac_tree, aux

    return jacfun



# -- Path setup --------------------------------------------------------------
import os
import sys
# autopep8: off
sys.path.insert(0, os.path.abspath('../'))
# must be called AFTER the above:
from grgrjax import __version__
# autopep8: on

# -- Project information -----------------------------------------------------
project = 'grgrjax'
copyright = '2022, Gregor Boehl'
author = 'Gregor Boehl'
version = __version__
release = version

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "nbsphinx",
    "sphinx.ext.mathjax",
    "IPython.sphinxext.ipython_console_highlighting",
]

templates_path = ["_templates", "**.ipynb_checkpoints"]
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------
html_theme = "sphinx_book_theme"
html_copy_source = True
html_show_sourcelink = True
html_sourcelink_suffix = ""
html_title = "grgrjax"
html_static_path = ["_static"]
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/gboehl/grgrjax",
    "repository_branch": "main",
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org",
        "notebook_interface": "classic",
    },
    "use_edit_page_button": True,
    "use_issues_button": True,
    "use_repository_button": True,
    "use_download_button": True,
}

autoclass_content = "both"
autodoc_member_order = "bysource"
latex_use_parts = False
master_doc = 'index'



# -*- coding: utf-8 -*-

import jax
import time
import jax.numpy as jnp
import scipy.sparse as ssp
from .helpers import val_and_jacfwd, amax

try:
    ssp_csr_array = ssp.csr_array
    ssp_lil_array = ssp.lil_array
except AttributeError:
    ssp_csr_array = ssp._arrays.csr_array
    ssp_lil_array = ssp._arrays.lil_array


def _newton_cond_func(carry):
    (xi, eps, cnt), (func, verbose, maxit, tol) = carry
    cond = cnt < maxit
    cond = jnp.logical_and(cond, eps > tol)
    cond = jnp.logical_and(cond, ~jnp.isnan(eps))
    verbose = jnp.logical_and(cnt, verbose)
    jax.debug.callback(callback_func, cnt, eps, verbose=verbose)
    return cond


def _newton_body_func(carry):
    (xi, eps, cnt), (func, verbose, maxit, tol) = carry
    xi_old = xi
    f, jac = func(xi)
    xi -= jax.scipy.linalg.solve(jac, f)
    eps = amax(xi-xi_old)
    return (xi, eps, cnt+1), (func, verbose, maxit, tol)


def callback_func(cnt, err, *args, fev=None, ltime=None, verbose=True):
    """Print a formatted on-line update for a iterative process.
    """
    mess = f'    Iteration {cnt:2d}'
    if fev is not None:
        mess += f' | fev. {fev:3d}'
    mess += f' | error {err:.2e}'
    for misc in args:
        mess += misc
    if ltime is not None:
        mess += f' | lapsed {ltime:3.4f}s'
    if verbose:
        print(mess)


@jax.jit
def newton_jax_jit(func, init, maxit=30, tol=1e-8, verbose=True):
    """Newton method for root finding of `func` using automatic differentiation with jax and running in and as jitted jax.

    Parameters
    ----------
    func : callable
        Function returning (y, jac) where f(x)=y=0 should be found and jac is the jacobian. Must be jittable with jax. Could e.g. be the output of jacfwd_and_val. The function must be a jax.
    init : array
        Initial values of x
    maxit : int, optional
        Maximum number of iterations
    tol : float, optional
        Required tolerance. Defaults to 1e-8

    Returns
    -------
    xopt : array
        Solution value x for f
    (fopt, jacopt): tuple of arrays
        : value (`fopt`) and Jacobian (`jacopt`) of `func` at `xopt`
    niter : int
        Number of iterations
    success: bool
        Wether the convergence criterion was reached
    """
    (xi, eps, cnt), _ = jax.lax.while_loop(_newton_cond_func,
                                           _newton_body_func, ((init, 1., 0), (func, verbose, maxit, tol)))
    return xi, func(xi), cnt, eps > tol


def _perform_checks_newton(res, eps, cnt, jac_is_nan, tol, rtol, maxit):

    if jac_is_nan.any():
        res['success'] = False
        res['message'] = "The Jacobian contains NaNs."
        return True

    if eps < tol or eps < rtol:
        res['success'] = True
        res['message'] = "The solution converged."
        return True

    if cnt == maxit:
        res['success'] = False
        res['message'] = f"Maximum number of {maxit} iterations reached."
        return True

    if jnp.isnan(eps):
        res['success'] = False
        res['message'] = f"Function returns NaNs"
        return True

    return False


def newton_jax(func, init, maxit=30, tol=1e-8, rtol=None, solver=None, verbose=True, verbose_jac=False):
    """Newton method for root finding of `func` using automatic differenciation with jax. The argument `func` must be jittable with jax. `newton_jax` itself is not jittable, for this use `newton_jax_jit`.

    Parameters
    ----------
    func : callable
        Function f for which f(x)=0 should be found. Is assumed to return a pair (value, jacobian) or (value, jacobian, aux). If not, `val_and_jacfwd` will be applied to the function, in which case the function must be jittable with jax.
    init : array
        Initial values of x
    maxit : int, optional
        Maximum number of iterations
    tol : float, optional
        Required tolerance. Defaults to 1e-8
    solver : callable, optional
        Provide a custom solver `solver(J,f)` for J@x = f. defaults to `jax.numpy.linalg.solve`
    verbose : bool, optional
        Whether to display messages
    verbose_jac : bool, optional
        Whether to supply additional information on the determinant of the jacobian (computationally more costly).

    Returns
    -------
    res: dict
        A dictionary of results similar to the output from scipy.optimize.root
    """

    st = time.time()
    verbose |= verbose_jac
    rtol = rtol or tol

    res = {}
    cnt = 0
    eps_fval = 1e8
    xi = jnp.array(init)

    while True:

        xold = xi.copy()
        cnt += 1
        # evaluate function
        fout = func(xi)

        # remap function if jacobian is not returned
        if not isinstance(fout, tuple):
            func = val_and_jacfwd(func)
            fout = func(xi)

        # unwrap results
        fval, jacval, aux = fout if len(fout) == 3 else (*fout, None)
        # check for convergence or errors
        jac_is_nan = jnp.isnan(jacval.data).any() if isinstance(
            jacval, ssp_csr_array) else jnp.isnan(jacval).any()
        eps = jnp.abs(fval).max()
        if _perform_checks_newton(res, eps, cnt, jac_is_nan, tol, rtol, maxit):
            break

        # be informative
        if verbose and cnt:
            ltime = time.time() - st
            info_str = f'    Iteration {cnt:3d} | max. error {eps:.2e} | lapsed {ltime:3.4f}'
            if verbose_jac:
                jacval = jacval.toarray() if isinstance(
                    jacval, ssp_csr_array) else jacval
                jacdet = jnp.linalg.det(jacval) if (
                    jacval.shape[0] == jacval.shape[1]) else 0
                info_str += f' | det {jacdet:1.5g} | rank {jnp.linalg.matrix_rank(jacval)}/{jacval.shape[0]}'
            print(info_str)

        # assign suitable solver if not given
        if solver is None:
            if isinstance(jacval, ssp_csr_array):
                solver = ssp.linalg.spsolve
            else:
                solver = jax.scipy.linalg.solve
        xi -= solver(jacval, fval)

    jacval = jacval.toarray() if isinstance(
        jacval, (ssp_csr_array, ssp_lil_array)) else jacval

    res['x'], res['niter'] = xi, cnt
    res['fun'], res['jac'] = fval, jacval
    if aux is not None:
        res['aux'] = aux

    if verbose_jac:
        # only calculate determinant if requested
        res['det'] = jnp.linalg.det(jacval) if (
            jacval.shape[0] == jacval.shape[1]) else 0
    else:
        res['det'] = None

    return res



# -*- coding: utf-8 -*-

import jax
import jax.numpy as jnp

from .__version__ import __version__
from .helpers import *
from .newton import *

__all__ = ["newton_jax", "newton_jax_jit", "amax", "callback_func", "jax_print",
           "jvp_vmap", "vjp_vmap", "val_and_jacfwd", "val_and_jacrev"]


def jax_print(w):
    """Print in jax compiled functions. Wrapper around `jax.experimental.host_callback.id_print`.
    """
    return jax.experimental.host_callback.id_print(w)



# -*- coding: utf-8 -*-

import jax
import jax.numpy as jnp
from . import newton_jax, newton_jax_jit, val_and_jacfwd


def f(x):
    res = x.copy()
    res = res.at[0].set(jnp.sin(x[0]))
    res = res.at[1].set(jnp.log(x[1]))
    return res


def solver(jval, fval):
    """A solver to solve indetermined problems.
    """
    return jnp.linalg.pinv(jval) @ fval


vaj_f = jax.tree_util.Partial(val_and_jacfwd(f))
x0 = jnp.ones(2)*.1


def test0():
    r0 = newton_jax(f, x0)
    assert r0['success']


def test1():
    r0 = newton_jax(vaj_f, x0)
    assert r0['success']


def test2():
    r0 = newton_jax(vaj_f, x0, verbose_jac=True)
    assert r0['success']


def test3():
    r0 = newton_jax(vaj_f, x0, solver=solver)
    assert r0['success']


def test4():
    r0 = newton_jax_jit(vaj_f, x0)
    assert not r0[-1]



from setuptools import setup, find_packages
from os import path

# get version from dedicated version file
version = {}
with open("grgrjax/__version__.py") as fp:
    exec(fp.read(), version)

# read the contents of the README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/gboehl/grgrjax",
    name='grgrjax',
    version=version['__version__'],
    author='Gregor Boehl',
    author_email='admin@gregorboehl.com',
    license='MIT',
    description='Some generic tools for JAX',
    classifiers=[
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(),
    install_requires=[
        'jax',
        'jaxlib',
    ],
    include_package_data=True,
)
